// This is free and unencumbered software released into the public
// domain.  For more information, see <http://unlicense.org> or the
// accompanying UNLICENSE file.

package main

import (
	"fmt"
	"os"
	"os/exec"
	"path/filepath"
	"reflect"
	"strings"
	"time"

	"git.sr.ht/~nelsam/hel/mock"
	"git.sr.ht/~nelsam/hel/parse"
	"git.sr.ht/~nelsam/hel/typ"
	"github.com/spf13/cobra"
)

var (
	cmd           *cobra.Command
	goimportsPath string
)

func init() {
	output, err := exec.Command("which", "goimports").Output()
	if err != nil {
		fmt.Println("Could not locate goimports: ", err.Error())
		fmt.Println("If goimports is not installed, please install it somewhere in your path.  " +
			"See https://godoc.org/golang.org/x/tools/cmd/goimports.")
		os.Exit(1)
	}
	goimportsPath = strings.TrimSpace(string(output))

	cmd = &cobra.Command{
		Use:   "hel",
		Short: "A mock generator for Go",
		Long: "hel is a simple mock generator.  The origin of the name is the Norse goddess, Hel, " +
			"who guards over the souls of those unworthy to enter Valhalla.  You can probably " +
			"guess how much I like mocks.",
		Run: func(cmd *cobra.Command, args []string) {
			if len(args) > 0 {
				fmt.Print("Invalid usage.  Help:\n\n")
				cmd.HelpFunc()(nil, nil)
				os.Exit(1)
			}
			packagePatterns, err := cmd.Flags().GetStringSlice("package")
			if err != nil {
				panic(err)
			}
			typePatterns, err := cmd.Flags().GetStringSlice("type")
			if err != nil {
				panic(err)
			}
			outputName, err := cmd.Flags().GetString("output")
			if err != nil {
				panic(err)
			}
			chanSize, err := cmd.Flags().GetInt("chan-size")
			if err != nil {
				panic(err)
			}
			noTestPkg, err := cmd.Flags().GetBool("no-test-package")
			if err != nil {
				panic(err)
			}
			local, err := cmd.Flags().GetBool("local")
			if err != nil {
				panic(err)
			}
			deps, err := cmd.Flags().GetBool("deps")
			if err != nil {
				panic(err)
			}

			fmt.Printf("Parsing directories matching %s %v", pluralize(packagePatterns, "pattern", "patterns"), packagePatterns)
			dirList := progress(func() []parse.Dir {
				return parse.Packages(packagePatterns...)
			})
			fmt.Print("\n")
			fmt.Println("Found directories:")
			for _, dir := range dirList {
				fmt.Println("  " + dir.Path())
			}
			fmt.Print("\n")

			fmt.Printf("Loading interface types in matching directories")
			typeDirs := progress(func() typ.Dirs {
				var godirs []typ.GoDir
				for _, dir := range dirList {
					godirs = append(godirs, dir)
				}
				return typ.Load(noTestPkg, godirs...).Filter(typePatterns...)
			})
			fmt.Print("\n\n")

			fmt.Printf("Generating mocks in output file %s", outputName)
			progress(func() struct{} {
				var opts []mock.Opt
				if local != deps {
					// This is odd and mostly related to backward compatibility,
					// but if all or none of the modes are set, it triggers the
					// same behavior. Only when they differ do we set options.
					if local {
						opts = append(opts, mock.ForLocalInterfaces())
					}
					if deps {
						opts = append(opts, mock.ForConcreteDependencies())
					}
				}
				for _, typeDir := range typeDirs.Slice() {
					mockPath, err := makeMocks(typeDir, outputName, chanSize, !noTestPkg, opts...)
					if err != nil {
						panic(err)
					}
					if mockPath != "" {
						if err = exec.Command(goimportsPath, "-w", mockPath).Run(); err != nil {
							panic(err)
						}
					}
				}
				return struct{}{}
			})
			fmt.Print("\n")
		},
	}
	cmd.Flags().StringSliceP("package", "p", []string{"."}, "The package(s) to generate mocks for.")
	cmd.Flags().StringSliceP("type", "t", []string{}, "The type(s) to generate mocks for.  If no types "+
		"are passed in, all exported interface types will be generated.")
	cmd.Flags().StringP("output", "o", "helheim_test.go", "The file to write generated mocks to.  Since hel does "+
		"not generate exported types, this file will be saved directly in all packages with generated mock.  "+
		"Also note that, since the types are not exported, you will want the file to end in '_test.go'.")
	cmd.Flags().IntP("chan-size", "s", 100, "The size of channels used for method calls.")
	cmd.Flags().Bool("no-test-package", false, "Generate mocks in the primary package rather than in {pkg}_test")
	cmd.Flags().BoolP("local", "l", false, "Generate mocks for local interface types only. If neither this nor --deps/-d are set, both will be enabled.")
	cmd.Flags().BoolP("deps", "d", false, "Generate mocks for all types used by concrete logic. If neither this nor --local/-l are set, both will be enabled.")
}

func makeMocks(types *typ.Dir, fileName string, chanSize int, useTestPkg bool, opts ...mock.Opt) (filePath string, err error) {
	mocks, err := mock.Generate(types, opts...)
	if err != nil {
		return "", fmt.Errorf("failed to generate mocks: %w", err)
	}
	if mocks.Len() == 0 {
		return "", nil
	}
	if useTestPkg {
		mocks.PrependLocalPackage(types.Package())
	}
	filePath = filepath.Join(types.Dir(), fileName)
	f, err := os.Create(filePath)
	if err != nil {
		return "", fmt.Errorf("failed to create output file: %w", err)
	}
	defer f.Close()
	testPkg := types.Package()
	if useTestPkg {
		testPkg += "_test"
	}
	if err := mocks.Output(testPkg, types.Dir(), chanSize, f); err != nil {
		return "", fmt.Errorf("failed to write mocks to output: %w", err)
	}
	return filePath, nil
}

func progress[T any](f func() T) T {
	stop := make(chan struct{})
	done := showProgress(stop)
	defer func() {
		close(stop)
		<-done
	}()

	return f()
}

func showProgress(stop <-chan struct{}) <-chan struct{} {
	done := make(chan struct{})
	go func() {
		defer close(done)
		ticker := time.NewTicker(time.Second / 2)
		defer ticker.Stop()
		for {
			select {
			case <-ticker.C:
				fmt.Print(".")
			case <-stop:
				return
			}
		}
	}()
	return done
}

type lengther interface {
	Len() int
}

func pluralize(values interface{}, singular, plural string) string {
	length := findLength(values)
	if length == 1 {
		return singular
	}
	return plural
}

func findLength(values interface{}) int {
	if lengther, ok := values.(lengther); ok {
		return lengther.Len()
	}
	return reflect.ValueOf(values).Len()
}

func main() {
	cmd.Execute()
}
