// This is free and unencumbered software released into the public
// domain.  For more information, see <http://unlicense.org> or the
// accompanying UNLICENSE file.

//go:generate hel --local

package mock

import (
	"bytes"
	"errors"
	"fmt"
	"go/ast"
	"go/format"
	"go/parser"
	"go/token"
	"io"
	"os"
	"strconv"

	"git.sr.ht/~nelsam/hel/typ"
)

const commentHeader = `// Code generated by git.sr.ht/~nelsam/hel. DO NOT EDIT.
//
// This file contains mocks generated by hel. Do not edit this code by
// hand unless you *really* know what you're doing. Expect any changes
// made manually to be overwritten the next time hel regenerates this
// file.

`

// TypeFinder represents a type which knows about types and dependencies.
type TypeFinder interface {
	Decls() []ast.Decl
	Dependencies(name string) []typ.Dependency
}

type import_ struct {
	name string
	path string
}

func (i import_) ident() *ast.Ident {
	if i.name == "" {
		return nil
	}
	return &ast.Ident{Name: i.name}
}

func (i import_) ast() *ast.ImportSpec {
	return &ast.ImportSpec{
		Name: i.ident(),
		Path: &ast.BasicLit{Kind: token.STRING, Value: strconv.Quote(i.path)},
	}
}

// Mocks is a set of mocks generated from a go package.
type Mocks struct {
	imports []import_
	mocks   []Mock
}

// Output writes the go code representing m to dest.  pkg will be the
// package name; dir is the destination directory (needed for formatting
// the file); chanSize is the buffer size of any channels created in
// constructors.
func (m *Mocks) Output(pkg, dir string, chanSize int, dest io.Writer) error {
	if _, err := dest.Write([]byte(commentHeader)); err != nil {
		return fmt.Errorf("mock: failed to write comment header: %w", err)
	}

	fset := token.NewFileSet()

	imps := []ast.Spec{
		&ast.ImportSpec{Path: &ast.BasicLit{Kind: token.STRING, Value: `"time"`}},
		&ast.ImportSpec{Path: &ast.BasicLit{Kind: token.STRING, Value: `"git.sr.ht/~nelsam/hel/vegr"`}},
	}
	for _, i := range m.imports {
		imps = append(imps, i.ast())
	}
	impDecl := &ast.GenDecl{Tok: token.IMPORT, Specs: imps}
	decls := []ast.Decl{impDecl}
	f := &ast.File{
		Name:  &ast.Ident{Name: pkg},
		Decls: append(decls, m.decls(chanSize)...),
	}

	var b bytes.Buffer
	if err := format.Node(&b, fset, f); err != nil {
		return fmt.Errorf("mock: failed to run initial formatting pass: %w", err)
	}

	fset = token.NewFileSet()
	file, err := parser.ParseFile(fset, pkg, &b, 0)
	if err != nil {
		out, tmpErr := os.CreateTemp("", "*.go")
		if tmpErr != nil {
			return fmt.Errorf("mock: failed to parse formatted code [could not save output for debugging]: %w", err)
		}
		out.WriteString(b.String())
		return fmt.Errorf("mock: failed to parse formatted code [saved for debugging as %v]: %w", out.Name(), err)
	}

	if err := format.Node(dest, fset, file); err != nil {
		return fmt.Errorf("mock: failed to run final formatting pass to file: %w", err)
	}
	return nil
}

// PrependLocalPackage prepends name as the package name for local types
// in m's signature.  This is most often used when mocking types that are
// imported by the local package.
func (m *Mocks) PrependLocalPackage(name string) {
	for i := 0; i < len(m.mocks); i++ {
		mock := m.mocks[i]
		if err := mock.PrependLocalPackage(name); err != nil {
			fmt.Printf("\nWARN: type %v could not prepend package selector, mocks will not be generated: %v\n", mock.Name(), err)
			m.mocks = append(m.mocks[:i], m.mocks[i+1:]...)
			i--
		}
	}
}

func (m *Mocks) decls(chanSize int) (decls []ast.Decl) {
	for _, mock := range m.mocks {
		decls = append(decls, mock.Ast(chanSize)...)
	}
	return decls
}

func (m *Mocks) Len() int {
	return len(m.mocks)
}

func (m *Mocks) Index(i int) Mock {
	return m.mocks[i]
}

func (m *Mocks) addImport(name, path string) error {
	for _, i := range m.imports {
		if i.path != path {
			continue
		}
		if i.name != name {
			return fmt.Errorf("mock: cannot generate mocks: conflicting names %v and %v for import %v", name, i.name, path)
		}
		return nil
	}
	m.imports = append(m.imports, import_{
		name: name,
		path: path,
	})
	return nil
}

type genMode int

const (
	modeLocalInterfaces = 1 << iota
	modeConcreteDeps

	modeAll = modeLocalInterfaces | modeConcreteDeps
)

type genPrefs struct {
	mode genMode
}

// Opt is an option function that can apply preferences for generating mocks.
type Opt func(genPrefs) genPrefs

// ForLocalInterfaces tells Generate to generate mocks for local interface
// types.
//
// If none of the For* options are included, the default will be to generate
// mocks for all strategies.
func ForLocalInterfaces() Opt {
	return func(p genPrefs) genPrefs {
		p.mode |= modeLocalInterfaces
		return p
	}
}

// ForConcreteDependencies tells Generate to generate mocks for any interface
// types that are depended on by functions, methods, or struct fields in
// exported code.
//
// If none of the For* options are included, the default will be to generate
// mocks for all strategies.
func ForConcreteDependencies() Opt {
	return func(p genPrefs) genPrefs {
		p.mode |= modeConcreteDeps
		return p
	}
}

// Generate generates a Mocks value for all interface types that are required,
// according to finder.
func Generate(finder TypeFinder, opts ...Opt) (*Mocks, error) {
	var prefs genPrefs
	for _, o := range opts {
		prefs = o(prefs)
	}
	if prefs.mode == 0 {
		prefs.mode = modeAll
	}
	base := finder.Decls()
	var (
		typs []*ast.TypeSpec
		deps []typ.Dependency
	)
	for _, typ := range base {
		switch typ := typ.(type) {
		case *ast.GenDecl:
			for _, spec := range typ.Specs {
				switch spec := spec.(type) {
				case *ast.TypeSpec:
					switch spec.Type.(type) {
					case *ast.InterfaceType:
						if prefs.mode&modeLocalInterfaces == 0 {
							continue
						}
					case *ast.StructType:
						if prefs.mode&modeConcreteDeps == 0 {
							continue
						}
					}
					typs = append(typs, spec)
					deps = append(deps, finder.Dependencies(spec.Name.Name)...)
				}
			}
		case *ast.FuncDecl:
			if prefs.mode&modeConcreteDeps == 0 {
				continue
			}
			deps = append(deps, finder.Dependencies(typ.Name.Name)...)
		}
	}
	deps = deDupe(typs, deps)
	var m Mocks
	for _, typ := range typs {
		inter, ok := typ.Type.(*ast.InterfaceType)
		if !ok {
			continue
		}
		var params []*ast.Field
		if typ.TypeParams != nil {
			params = typ.TypeParams.List
		}
		newMock, err := For(typ.Name.Name, params, inter)
		if errors.Is(err, ErrUnexported) {
			fmt.Printf("\nWARN: could not generate a mock for %v: %v\n", typ.Name.Name, err)
			continue
		}
		if err != nil {
			return nil, fmt.Errorf("mock: failed to generate mocks: %w", err)
		}
		m.mocks = append(m.mocks, newMock)
	}
	for _, dep := range deps {
		pkgName := dep.PkgName()
		if !dep.ImportAliased() {
			pkgName = ""
		}
		if err := m.addImport(pkgName, dep.PkgPath()); err != nil {
			return nil, fmt.Errorf("mock: failed to generate dependency mock %v: %w", dep.Name(), err)
		}
		inter, ok := dep.Type().(*ast.InterfaceType)
		if !ok {
			continue
		}
		newMock, err := For(dep.TypeName(), dep.Params(), inter)
		if errors.Is(err, ErrUnexported) {
			fmt.Printf("\nWARN: could not generate a mock for dependency type %v: %v\n", dep.Name(), err)
			continue
		}
		if err != nil {
			return nil, fmt.Errorf("mock: failed to generate dependency mock %v: %w", dep.Name(), err)
		}
		selectorErr := newMock.PrependLocalPackage(dep.PkgName())
		if errors.Is(selectorErr, ErrUnexported) {
			fmt.Printf("\nWARN: could not generate a mock for dependency type %v: type signature contains unexported types: %v\n", dep.Name(), selectorErr)
			continue
		}
		if selectorErr != nil {
			return nil, fmt.Errorf("mock: failed to add package selector to dependency mock %v: %w", dep.Name(), selectorErr)
		}
		m.mocks = append(m.mocks, newMock)
	}
	return &m, nil
}

func deDupe(typs []*ast.TypeSpec, deps []typ.Dependency) []typ.Dependency {
	for _, typ := range typs {
		for i := 0; i < len(deps); i++ {
			if deps[i].TypeName() != typ.Name.Name {
				continue
			}
			if deps[i].PkgPath() == "" {
				deps = append(deps[:i], deps[i+1:]...)
				i--
				continue
			}
			deps[i] = deps[i].AvoidCollision(typ.Name.Name)
		}
	}
	for i := 0; i < len(deps); i++ {
		for j := i + 1; j < len(deps); j++ {
			if equal(deps[i], deps[j]) {
				deps = append(deps[:j], deps[j+1:]...)
				j--
				continue
			}
			deps[j] = deps[j].AvoidCollision(deps[i].TypeName())
		}
	}
	return deps
}

func equal(a, b typ.Dependency) bool {
	if a.PkgPath() != b.PkgPath() {
		return false
	}
	if a.TypeName() != b.TypeName() {
		return false
	}
	return true
}
